@@ -6,16 +6,19 @@ import collections
 import logging
 import threading
 import operator
+import warnings
 
 from pymorphy2 import opencorpora_dict
-from pymorphy2 import units
 from pymorphy2.dawg import ConditionalProbDistDAWG
+import pymorphy2.lang
 
 logger = logging.getLogger(__name__)
 
 _Parse = collections.namedtuple('Parse', 'word, tag, normal_form, score, methods_stack')
 
 _score_getter = operator.itemgetter(3)
+auto = object()
+
 
 class Parse(_Parse):
     """
@@ -46,7 +49,10 @@ class Parse(_Parse):
     @property
     def is_known(self):
         """ True if this form is a known dictionary form. """
-        return self._dict.word_is_known(self.word, strict_ee=True)
+        return self._dict.word_is_known(
+            word=self.word,
+            substitutes_compiled=self._morph.char_substitutes
+        )
 
     @property
     def normalized(self):
@@ -59,7 +65,7 @@ class Parse(_Parse):
     #     return self._dict.build_paradigm_info(self.para_id)
 
 
-class SingleTagProbabilityEstimator(object):
+class ProbabilityEstimator(object):
     def __init__(self, dict_path):
         cpd_path = os.path.join(dict_path, 'p_t_given_w.intdawg')
         self.p_t_given_w = ConditionalProbDistDAWG().load(cpd_path)
@@ -95,15 +101,34 @@ class SingleTagProbabilityEstimator(object):
         )
 
 
-class DummySingleTagProbabilityEstimator(object):
-    def __init__(self, dict_path):
+def _lang_dict_paths():
+    import pkg_resources
+    paths = dict(
+        (pkg.name, pkg.load().get_path())
+        for pkg in pkg_resources.iter_entry_points('pymorphy2_dicts')
+    )
+
+    try:
+        import pymorphy2_dicts
+        paths['ru-old'] = pymorphy2_dicts.get_path()
+    except ImportError:
         pass
 
-    def apply_to_parses(self, word, word_lower, parses):
-        return parses
+    return paths
 
-    def apply_to_tags(self, word, word_lower, tags):
-        return tags
+
+def lang_dict_path(lang):
+    """ Return language-specific dictionary path """
+    lang_paths = _lang_dict_paths()
+    if lang in lang_paths:
+        return lang_paths[lang]
+
+    raise ValueError(
+        "Can't find a dictionary for language %r. Installed languages: %r. "
+        "Try installing pymorphy2-dicts-%s package." % (
+            lang, list(lang_paths.keys()), lang
+        )
+    )
 
 
 class MorphAnalyzer(object):
@@ -130,7 +155,7 @@ class MorphAnalyzer(object):
     with a path to dictionaries, or pass ``path`` argument
     to :class:`pymorphy2.MorphAnalyzer` constructor::
 
-        >>> morph = pymorphy2.MorphAnalyzer('/path/to/dictionaries') # doctest: +SKIP
+        >>> morph = pymorphy2.MorphAnalyzer(path='/path/to/dictionaries') # doctest: +SKIP
 
     By default, methods of this class return parsing results
     as namedtuples :class:`Parse`. This has performance implications
@@ -140,41 +165,32 @@ class MorphAnalyzer(object):
         >>> morph = pymorphy2.MorphAnalyzer(result_type=None)
 
     """
+    DICT_PATH_ENV_VARIABLE = 'PYMORPHY2_DICT_PATH'
+    DEFAULT_UNITS = pymorphy2.lang.ru.DEFAULT_UNITS
+    DEFAULT_SUBSTITUTES = pymorphy2.lang.ru.CHAR_SUBSTITUTES
+    char_substitutes = None
 
-    ENV_VARIABLE = 'PYMORPHY2_DICT_PATH'
-    DEFAULT_UNITS = [
-        [
-            units.DictionaryAnalyzer,
-            units.AbbreviatedFirstNameAnalyzer,
-            units.AbbreviatedPatronymicAnalyzer,
-        ],
+    _lock = threading.RLock()
 
-        units.NumberAnalyzer,
-        units.PunctuationAnalyzer,
-        [
-            units.RomanNumberAnalyzer,
-            units.LatinAnalyzer
-        ],
+    def __init__(self, path=None, lang=None, result_type=Parse, units=None,
+                 probability_estimator_cls=auto, char_substitutes=auto):
 
-        units.HyphenSeparatedParticleAnalyzer,
-        units.HyphenAdverbAnalyzer,
-        units.HyphenatedWordsAnalyzer,
-        units.KnownPrefixAnalyzer,
-        [
-            units.UnknownPrefixAnalyzer,
-            units.KnownSuffixAnalyzer
-        ],
-        units.UnknAnalyzer,
-    ]
+        # save arguments for pickling/unpickling
+        self._path = path
+        self._lang = lang
 
-    def __init__(self, path=None, result_type=Parse, units=None,
-                 probability_estimator_cls=SingleTagProbabilityEstimator):
-        path = self.choose_dictionary_path(path)
-        with threading.RLock():
+        if path is None and lang is None:
+            lang = 'ru'
+
+        path = self.choose_dictionary_path(path, lang)
+
+        with self._lock:
             self.dictionary = opencorpora_dict.Dictionary(path)
-            if probability_estimator_cls is None:
-                probability_estimator_cls = DummySingleTagProbabilityEstimator
-            self.prob_estimator = probability_estimator_cls(path)
+            self.lang = self.choose_language(self.dictionary, lang)
+
+            self.prob_estimator = self._get_prob_estimator(
+                probability_estimator_cls, self.dictionary, path
+            )
 
             if result_type is not None:
                 # create a subclass with the same name,
@@ -189,39 +205,83 @@ class MorphAnalyzer(object):
                 self._result_type = None
 
             self._result_type_orig = result_type
+            self._init_char_substitutes(char_substitutes)
             self._init_units(units)
 
-    def _init_units(self, unit_classes=None):
-        if unit_classes is None:
-            unit_classes = self.DEFAULT_UNITS
+    def _init_units(self, units_unbound=None):
+        if units_unbound is None:
+            units_unbound = self._config_value('DEFAULT_UNITS', self.DEFAULT_UNITS)
 
-        self._unit_classes = unit_classes
+        self._units_unbound = units_unbound
         self._units = []
-        for item in unit_classes:
+        for item in units_unbound:
             if isinstance(item, (list, tuple)):
-                for cls in item[:-1]:
-                    self._units.append((cls(self), False))
-                self._units.append((item[-1](self), True))
+                for unit in item[:-1]:
+                    self._units.append((self._bound_unit(unit), False))
+                self._units.append((self._bound_unit(item[-1]), True))
             else:
-                self._units.append((item(self), True))
+                self._units.append((self._bound_unit(item), True))
+
+    def _init_char_substitutes(self, char_substitutes):
+        if char_substitutes is auto:
+            char_substitutes = self._config_value('CHAR_SUBSTITUTES', self.DEFAULT_SUBSTITUTES)
+        self.char_substitutes = self.dictionary.words.compile_replaces(char_substitutes or {})
+
+    def _bound_unit(self, unit):
+        unit = unit.clone()
+        unit.init(self)
+        return unit
+
+    def _lang_default_config(self):
+        assert self.lang is not None
+        aliases = {'ru-old': 'ru'}
+        lang = aliases.get(self.lang, self.lang)
+        if not hasattr(pymorphy2.lang, lang):
+            warnings.warn("unknown language code: %r" % lang)
+            return None
+        return getattr(pymorphy2.lang, lang)
+
+    def _config_value(self, key, default):
+        config = self._lang_default_config()
+        return getattr(config, key, default)
 
     @classmethod
-    def choose_dictionary_path(cls, path=None):
+    def _get_prob_estimator(cls, estimator_cls, dictionary, path):
+        if estimator_cls is auto:
+            if dictionary.meta.get('P(t|w)'):
+                estimator_cls = ProbabilityEstimator
+        if estimator_cls is auto or estimator_cls is None:
+            return None
+        return estimator_cls(path)
+
+    @classmethod
+    def choose_dictionary_path(cls, path=None, lang=None):
         if path is not None:
             return path
 
-        if cls.ENV_VARIABLE in os.environ:
-            return os.environ[cls.ENV_VARIABLE]
+        if cls.DICT_PATH_ENV_VARIABLE in os.environ:
+            return os.environ[cls.DICT_PATH_ENV_VARIABLE]
 
-        try:
-            import pymorphy2_dicts
-            return pymorphy2_dicts.get_path()
-        except ImportError:
-            msg = ("Can't find dictionaries. "
-                   "Please either pass a path to dictionaries, "
-                   "or install 'pymorphy2-dicts' package, "
-                   "or set %s environment variable.") % cls.ENV_VARIABLE
-            raise ValueError(msg)
+        return lang_dict_path(lang)
+
+    @classmethod
+    def choose_language(cls, dictionary, lang):
+        if lang is None:
+            if dictionary.lang is None:
+                # this could be e.g. old pymorphy2 dictionary
+                warnings.warn("Dictionary doesn't declare its language; "
+                              "assuming 'ru'")
+                return 'ru'
+            return dictionary.lang
+
+        if dictionary.lang != lang:
+            # allow incorrect 'lang' values, but show a warning
+            warnings.warn(
+                "Dictionary language (%r) doesn't match "
+                "analyzer language (%r)." % (dictionary.lang, lang)
+            )
+
+        return lang
 
     def parse(self, word):
         """
@@ -242,7 +302,8 @@ class MorphAnalyzer(object):
             if is_terminal and res:
                 break
 
-        res = self.prob_estimator.apply_to_parses(word, word_lower, res)
+        if self.prob_estimator is not None:
+            res = self.prob_estimator.apply_to_parses(word, word_lower, res)
 
         if self._result_type is None:
             return res
@@ -260,7 +321,9 @@ class MorphAnalyzer(object):
             if is_terminal and res:
                 break
 
-        return self.prob_estimator.apply_to_tags(word, word_lower, res)
+        if self.prob_estimator is not None:
+            res = self.prob_estimator.apply_to_tags(word, word_lower, res)
+        return res
 
     def normal_forms(self, word):
         """
@@ -302,7 +365,7 @@ class MorphAnalyzer(object):
         grammemes = form[1].updated_grammemes(required_grammemes)
         def similarity(frm):
             tag = frm[1]
-            return len(grammemes & tag.grammemes)
+            return len(grammemes & tag.grammemes) - 0.1 * len(grammemes ^ tag.grammemes)
 
         return heapq.nlargest(1, possible_results, key=similarity)
 
@@ -324,11 +387,14 @@ class MorphAnalyzer(object):
             else:
                 yield self._result_type(*parse)
 
-    def word_is_known(self, word, strict_ee=False):
+    def word_is_known(self, word, strict=False):
         """
         Check if a ``word`` is in the dictionary.
-        Pass ``strict_ee=True`` if ``word`` is guaranteed to
-        have correct е/ё letters.
+
+        By default, some fuzziness is allowed, depending on a
+        dictionary - e.g. for Russian ё letters replaced with е are handled.
+        Pass ``strict=True`` to make matching strict (e.g. if it is
+        guaranteed the ``word`` has correct е/ё or г/ґ letters).
 
         .. note::
 
@@ -338,7 +404,10 @@ class MorphAnalyzer(object):
             method should be used with extra care.
 
         """
-        return self.dictionary.word_is_known(word.lower(), strict_ee)
+        return self.dictionary.word_is_known(
+            word = word.lower(),
+            substitutes_compiled = None if strict else self.char_substitutes
+        )
 
     @property
     def TagClass(self):
@@ -356,7 +425,5 @@ class MorphAnalyzer(object):
         return self.TagClass.lat2cyr(tag_or_grammeme)
 
     def __reduce__(self):
-        args = (self.dictionary.path, self._result_type_orig, self._unit_classes)
+        args = (self._path, self._lang, self._result_type_orig, self._units_unbound)
         return self.__class__, args, None
-
-
