@@ -18,20 +18,25 @@ except AttributeError:
     izip = zip
 
 from pymorphy2 import dawg
-from pymorphy2.constants import PARADIGM_PREFIXES, PREDICTION_PREFIXES
-from pymorphy2.utils import longest_common_substring, largest_group
+from pymorphy2.utils import (
+    longest_common_substring,
+    largest_elements,
+    with_progress
+)
 
 logger = logging.getLogger(__name__)
 
 
 CompiledDictionary = collections.namedtuple(
     'CompiledDictionary',
-    'gramtab suffixes paradigms words_dawg prediction_suffixes_dawgs parsed_dict prediction_options'
+    'gramtab suffixes paradigms words_dawg prediction_suffixes_dawgs parsed_dict compile_options'
 )
 
+_pick_second_item = operator.itemgetter(1)
 
-def convert_to_pymorphy2(opencorpora_dict_path, out_path, overwrite=False,
-                         prediction_options=None):
+
+def convert_to_pymorphy2(opencorpora_dict_path, out_path, source_name,
+                         language_code, overwrite=False, compile_options=None):
     """
     Convert a dictionary from OpenCorpora XML format to
     Pymorphy2 compacted format.
@@ -39,7 +44,7 @@ def convert_to_pymorphy2(opencorpora_dict_path, out_path, overwrite=False,
     ``out_path`` should be a name of folder where to put dictionaries.
     """
     from .parse import parse_opencorpora_xml
-    from .preprocess import simplify_tags
+    from .preprocess import simplify_tags, drop_unsupported_parses
     from .storage import save_compiled_dict
 
     dawg.assert_can_create()
@@ -48,54 +53,57 @@ def convert_to_pymorphy2(opencorpora_dict_path, out_path, overwrite=False,
 
     parsed_dict = parse_opencorpora_xml(opencorpora_dict_path)
     simplify_tags(parsed_dict)
-    compiled_dict = compile_parsed_dict(parsed_dict, prediction_options)
-    save_compiled_dict(compiled_dict, out_path)
+    drop_unsupported_parses(parsed_dict)
+    compiled_dict = compile_parsed_dict(parsed_dict, compile_options)
+    save_compiled_dict(compiled_dict, out_path,
+                       source_name=source_name, language_code=language_code)
 
 
-def compile_parsed_dict(parsed_dict, prediction_options=None):
+def compile_parsed_dict(parsed_dict, compile_options=None):
     """
     Return compacted dictionary data.
     """
-    _prediction_options = dict(
-        # defaults
+    options = dict(
         min_ending_freq=2,
         min_paradigm_popularity=3,
-        max_suffix_length=5
+        max_suffix_length=5,
     )
-    _prediction_options.update(prediction_options or {})
+    options.update(compile_options or {})
+    paradigm_prefixes = options["paradigm_prefixes"]
 
     gramtab = []
     paradigms = []
     words = []
 
-    seen_tags = dict()
-    seen_paradigms = dict()
+    tag_ids = dict()
+    paradigm_ids = dict()
 
-    logger.info("inlining lexeme derivational rules...")
+    logger.info("inlining lexeme derivational rules")
     lexemes = _join_lexemes(parsed_dict.lexemes, parsed_dict.links)
+    logger.info("lexemes after link inlining: %s", len(lexemes))
 
-    logger.info('building paradigms...')
+    logger.info('building paradigms')
     logger.debug("%20s %15s %15s %15s", "word", "len(gramtab)", "len(words)", "len(paradigms)")
 
     paradigm_popularity = collections.defaultdict(int)
 
     for index, lexeme in enumerate(lexemes):
-        stem, paradigm = _to_paradigm(lexeme)
+        stem, paradigm = _to_paradigm(lexeme, paradigm_prefixes)
 
         # build gramtab
         for suff, tag, pref in paradigm:
-            if tag not in seen_tags:
-                seen_tags[tag] = len(gramtab)
+            if tag not in tag_ids:
+                tag_ids[tag] = len(gramtab)
                 gramtab.append(tag)
 
         # build paradigm index
-        if paradigm not in seen_paradigms:
-            seen_paradigms[paradigm] = len(paradigms)
+        if paradigm not in paradigm_ids:
+            paradigm_ids[paradigm] = len(paradigms)
             paradigms.append(
-                tuple([(suff, seen_tags[tag], pref) for suff, tag, pref in paradigm])
+                tuple([(suff, tag_ids[tag], pref) for suff, tag, pref in paradigm])
             )
 
-        para_id = seen_paradigms[paradigm]
+        para_id = paradigm_ids[paradigm]
         paradigm_popularity[para_id] += 1
 
         for idx, (suff, tag, pref) in enumerate(paradigm):
@@ -110,24 +118,27 @@ def compile_parsed_dict(parsed_dict, prediction_options=None):
 
 
     logger.debug("%20s %15s %15s %15s", "total:", len(gramtab), len(words), len(paradigms))
-    logger.debug("linearizing paradigms..")
+    logger.debug("linearizing paradigms")
 
     def get_form(para):
         return list(next(izip(*para)))
 
     forms = [get_form(para) for para in paradigms]
     suffixes = sorted(set(list(itertools.chain(*forms))))
-    suffixes_dict = dict(
+    suffix_ids = dict(
         (suff, index)
         for index, suff in enumerate(suffixes)
     )
 
+    paradigm_prefix_ids = dict(
+        (pref, idx) for idx, pref in enumerate(paradigm_prefixes)
+    )
     def fix_strings(paradigm):
         """ Replace suffix and prefix with the respective id numbers. """
         para = []
         for suff, tag, pref in paradigm:
             para.append(
-                (suffixes_dict[suff], tag, PARADIGM_PREFIXES.index(pref))
+                (suffix_ids[suff], tag, paradigm_prefix_ids[pref])
             )
         return para
 
@@ -136,22 +147,36 @@ def compile_parsed_dict(parsed_dict, prediction_options=None):
 
     logger.debug('calculating prediction data..')
     suffixes_dawgs_data = _suffixes_prediction_data(
-        words, paradigm_popularity, gramtab, paradigms, suffixes, **_prediction_options
+        words=words,
+        paradigm_popularity=paradigm_popularity,
+        gramtab=gramtab,
+        paradigms=paradigms,
+        suffixes=suffixes,
+        min_ending_freq=options["min_ending_freq"],
+        min_paradigm_popularity=options["min_paradigm_popularity"],
+        max_suffix_length=options["max_suffix_length"],
+        paradigm_prefixes=paradigm_prefixes,
     )
 
-    logger.debug('building word DAWG..')
+    logger.debug('building word DAFSA')
     words_dawg = dawg.WordsDawg(words)
 
     del words
 
     prediction_suffixes_dawgs = []
     for prefix_id, dawg_data in enumerate(suffixes_dawgs_data):
-        logger.debug('building prediction_suffixes DAWGs #%d..' % prefix_id)
+        logger.debug('building prediction_suffixes DAFSA #%d' % prefix_id)
         prediction_suffixes_dawgs.append(dawg.PredictionSuffixesDAWG(dawg_data))
 
-    return CompiledDictionary(tuple(gramtab), suffixes, paradigms,
-                              words_dawg, prediction_suffixes_dawgs,
-                              parsed_dict, _prediction_options)
+    return CompiledDictionary(
+        gramtab=tuple(gramtab),
+        suffixes=suffixes,
+        paradigms=paradigms,
+        words_dawg=words_dawg,
+        prediction_suffixes_dawgs=prediction_suffixes_dawgs,
+        parsed_dict=parsed_dict,
+        compile_options=options,
+    )
 
 
 def _join_lexemes(lexemes, links):
@@ -180,10 +205,17 @@ def _join_lexemes(lexemes, links):
 #    <type id="18">ADJF-SUPR_suppl</type>
 #    <type id="19">ADJF-SUPR_nai</type>
 #    <type id="20">ADJF-SUPR_slng</type>
+#    <type id="21">FULL-CONTRACTED</type>
+#    <type id="22">NORM-ORPHOVAR</type>
+#    <type id="23">CARDINAL-ORDINAL</type>      # e.g. первый - один
+#    <type id="24">SBST_MASC-SBST_FEMN</type>   # e.g. подсудимый - подсудимая
+#    <type id="25">SBST_MASC-SBST_PLUR</type>   # e.g. подсудимый - подсудимые
+#    <type id="26">ADVB-COMP</type>             # ??
+#    <type id="27">ADJF_TEXT-ADJF_NUMBER</type> # e.g. первый - 1-й
 #    </link_types>
 
-    EXCLUDED_LINK_TYPES = set([7, ])
-#    ALLOWED_LINK_TYPES = set([3, 4, 5])
+    EXCLUDED_LINK_TYPES = set(['7', '21', '23', '27'])
+#    ALLOWED_LINK_TYPES = set(['3', '4', '5'])
 
     moves = dict()
 
@@ -210,23 +242,27 @@ def _join_lexemes(lexemes, links):
     return [lexemes[lex_id] for lex_id in lex_ids if lexemes[lex_id]]
 
 
-def _to_paradigm(lexeme):
+def _to_paradigm(lexeme, paradigm_prefixes):
     """
     Extract (stem, paradigm) pair from lexeme (which is a list of
     (word_form, tag) tuples). Paradigm is a list of suffixes with
     associated tags and prefixes.
     """
     forms, tags = list(zip(*lexeme))
-    prefixes = [''] * len(tags)
 
     if len(forms) == 1:
         stem = forms[0]
+        prefixes = ['']
     else:
         stem = longest_common_substring(forms)
         prefixes = [form[:form.index(stem)] for form in forms]
 
         # only allow prefixes from PARADIGM_PREFIXES
-        if any(pref not in PARADIGM_PREFIXES for pref in prefixes):
+        if any(pref not in paradigm_prefixes for pref in prefixes):
+            # With right PARADIGM_PREFIXES empty stem is fine;
+            # os.path.commonprefix doesn't return anything useful
+            # for prediction.
+            # stem = os.path.commonprefix(forms)
             stem = ""
             prefixes = [''] * len(tags)
 
@@ -234,19 +270,15 @@ def _to_paradigm(lexeme):
         form[len(pref)+len(stem):]
         for form, pref in zip(forms, prefixes)
     )
-
     return stem, tuple(zip(suffixes, tags, prefixes))
 
 
 def _suffixes_prediction_data(words, paradigm_popularity, gramtab, paradigms, suffixes,
-                              min_ending_freq, min_paradigm_popularity, max_suffix_length):
+                              min_ending_freq, min_paradigm_popularity, max_suffix_length,
+                              paradigm_prefixes):
 
     logger.debug('calculating prediction data: removing non-productive paradigms..')
-    productive_paradigms = set(
-        para_id
-        for (para_id, count) in paradigm_popularity.items()
-        if count >= min_paradigm_popularity
-    )
+    productive_paradigms = _popular_keys(paradigm_popularity, min_paradigm_popularity)
 
     # ["suffix"] => number of occurrences
     # this is for removing non-productive suffixes
@@ -254,14 +286,14 @@ def _suffixes_prediction_data(words, paradigm_popularity, gramtab, paradigms, su
 
     # [form_prefix_id]["suffix"]["POS"][(para_id, idx)] => number or occurrences
     # this is for selecting most popular parses
-    endings = {}
-    for form_prefix_id in range(len(PARADIGM_PREFIXES)):
-        endings[form_prefix_id] = collections.defaultdict(
+    prefix_endings = {}
+    for form_prefix_id in range(len(paradigm_prefixes)):
+        prefix_endings[form_prefix_id] = collections.defaultdict(
                                     lambda: collections.defaultdict(
                                         lambda: collections.defaultdict(int)))
 
     logger.debug('calculating prediction data: checking word endings..')
-    for word, (para_id, idx) in words:
+    for word, (para_id, idx) in with_progress(words, "Checking word endings"):
 
         if para_id not in productive_paradigms:
             continue
@@ -272,7 +304,7 @@ def _suffixes_prediction_data(words, paradigm_popularity, gramtab, paradigms, su
 
         tag = gramtab[paradigm[form_count + idx]]
         form_prefix_id = paradigm[2*form_count + idx]
-        form_prefix = PARADIGM_PREFIXES[form_prefix_id]
+        form_prefix = paradigm_prefixes[form_prefix_id]
         form_suffix = suffixes[paradigm[idx]]
 
         assert len(word) >= len(form_prefix+form_suffix), word
@@ -287,38 +319,51 @@ def _suffixes_prediction_data(words, paradigm_popularity, gramtab, paradigms, su
 
         for i in range(max(len(form_suffix), 1), max_suffix_length+1): #was: 1,2,3,4,5
             word_end = word[-i:]
-
             ending_counts[word_end] += 1
-            endings[form_prefix_id][word_end][POS][(para_id, idx)] += 1
+            prefix_endings[form_prefix_id][word_end][POS][(para_id, idx)] += 1
 
     dawgs_data = []
 
-    for form_prefix_id in sorted(endings.keys()):
-        logger.debug('calculating prediction data: preparing DAWGs data #%d..' % form_prefix_id)
-        counted_suffixes_dawg_data = []
-        endings_with_prefix = endings[form_prefix_id]
-
-        for suff in endings_with_prefix:
-            if ending_counts[suff] < min_ending_freq:
-                continue
-
-            for POS in endings_with_prefix[suff]:
-
-                common_endings = largest_group(
-                    endings_with_prefix[suff][POS].items(),
-                    operator.itemgetter(1)
-                )
-
-                for form, cnt in common_endings:
-                    counted_suffixes_dawg_data.append(
-                        (suff, (cnt,) + form)
-                    )
-
-        dawgs_data.append(counted_suffixes_dawg_data)
+    for form_prefix_id in sorted(prefix_endings.keys()):
+        logger.debug('calculating prediction data: preparing DAFSA #%d..' % form_prefix_id)
+        endings = prefix_endings[form_prefix_id]
+        dawgs_data.append(
+            _get_suffixes_dawg_data(endings, ending_counts, min_ending_freq)
+        )
 
     return dawgs_data
 
 
+def _get_suffixes_dawg_data(endings, ending_counts, min_ending_freq):
+    counted_suffixes_dawg_data = []
+
+    for ending in endings:
+        if ending_counts[ending] < min_ending_freq:
+            continue
+
+        for POS in endings[ending]:
+
+            common_form_counts = largest_elements(
+                iterable=endings[ending][POS].items(),
+                key=_pick_second_item,
+                n=1,
+            )
+
+            for form, cnt in common_form_counts:
+                # form is a `(para_id, idx)` tuple here
+                # XXX: shouldn't we use inverted cnt to make the results
+                # sorted high to low?
+                counted_suffixes_dawg_data.append(
+                    (ending, (cnt,) + form)
+                )
+
+    return counted_suffixes_dawg_data
+
+
+def _popular_keys(counter, threshold):
+    return set(key for (key, count) in counter.items() if count >= threshold)
+
+
 def _linearized_paradigm(paradigm):
     """
     Convert ``paradigm`` (a list of tuples with numbers)
