@@ -25,12 +25,26 @@ except ImportError:
     def xml_clear_elem(elem):
         elem.clear()
 
+from pymorphy2.utils import with_progress
+
 
 logger = logging.getLogger(__name__)
 
 ParsedDictionary = collections.namedtuple('ParsedDictionary', 'lexemes links grammemes version revision')
 
 
+def get_dictionary_info(filename, elem_limit=1000):
+    """ Return dictionary version and revision """
+    for idx, (ev, elem) in enumerate(iterparse(filename, events=(str('start'),))):
+        if elem.tag == 'dictionary':
+            version = elem.get('version')
+            revision = elem.get('revision')
+            return version, revision
+        if idx > elem_limit:
+            return None, None
+    return None, None
+
+
 def parse_opencorpora_xml(filename):
     """
     Parse OpenCorpora dict XML and return a ``ParsedDictionary`` namedtuple.
@@ -39,19 +53,20 @@ def parse_opencorpora_xml(filename):
     links = []
     lexemes = {}
     grammemes = []
-    version, revision = None, None
-    _lexemes_len = 0
 
-    for ev, elem in iterparse(filename, events=(str('start'), str('end'))):
+    version, revision = get_dictionary_info(filename)
+    logger.info("dictionary v%s, rev%s", version, revision)
+    interesting_tags = set(['grammeme', 'lemma', 'link'])
 
-        if ev == 'start':
-            if elem.tag == 'dictionary':
-                version = elem.get('version')
-                revision = elem.get('revision')
-                logger.info("dictionary v%s, rev%s", version, revision)
-                xml_clear_elem(elem)
-            continue
+    def _parse(filename):
+        for ev, elem in iterparse(filename):
+            if elem.tag not in interesting_tags:
+                continue
+            yield ev, elem
 
+    logger.info("parsing XML dictionary")
+
+    for ev, elem in with_progress(_parse(filename), "XML parsing"):
         if elem.tag == 'grammeme':
             name = elem.find('name').text
             parent = elem.get('parent')
@@ -63,17 +78,11 @@ def parse_opencorpora_xml(filename):
             xml_clear_elem(elem)
 
         if elem.tag == 'lemma':
-            if not lexemes:
-                logger.info('parsing xml:lemmas...')
-
             lex_id, word_forms = _word_forms_from_xml_elem(elem)
             lexemes[lex_id] = word_forms
             xml_clear_elem(elem)
 
         elif elem.tag == 'link':
-            if not links:
-                logger.info('parsing xml:links...')
-
             link_tuple = (
                 elem.get('from'),
                 elem.get('to'),
@@ -82,15 +91,17 @@ def parse_opencorpora_xml(filename):
             links.append(link_tuple)
             xml_clear_elem(elem)
 
-        if len(lexemes) != _lexemes_len and not (len(lexemes) % 50000):
-            logger.debug("%d lexemes parsed" % len(lexemes))
-            _lexemes_len = len(lexemes)
-
-    return ParsedDictionary(lexemes, links, grammemes, version, revision)
+    return ParsedDictionary(
+        lexemes=lexemes,
+        links=links,
+        grammemes=grammemes,
+        version=version,
+        revision=revision
+    )
 
 
 def _grammemes_from_elem(elem):
-    return ",".join(g.get('v') for g in elem.findall('g'))
+    return ",".join([g.get('v') for g in elem.iter('g')])
 
 
 def _word_forms_from_xml_elem(elem):
@@ -103,16 +114,21 @@ def _word_forms_from_xml_elem(elem):
     if len(elem) == 0:  # deleted lexeme?
         return lex_id, lexeme
 
-    base_info = elem.findall('l')
+    base_info = list(elem.iter('l'))
 
     assert len(base_info) == 1
     base_grammemes = _grammemes_from_elem(base_info[0])
 
-    for form_elem in elem.findall('f'):
+    for form_elem in elem.iter('f'):
         grammemes = _grammemes_from_elem(form_elem)
         form = form_elem.get('t').lower()
+        if not (base_grammemes + grammemes):
+            logger.warning("no information provided for word %s, dropping the whole lexeme" % form)
+            return lex_id, []
+        if isinstance(form, bytes):  # Python 2.x
+            form = form.decode('ascii')
         lexeme.append(
-            (form, " ".join([base_grammemes, grammemes]).strip())
+            (form, (base_grammemes + " " + grammemes).strip())
         )
 
     return lex_id, lexeme
