@@ -25,14 +25,13 @@ class Dictionary(object):
         self.paradigm_prefixes = self._data.paradigm_prefixes
         self.suffixes = self._data.suffixes
         self.words = self._data.words
-        self.prediction_prefixes = self._data.prediction_prefixes
         self.prediction_suffixes_dawgs = self._data.prediction_suffixes_dawgs
         self.meta = self._data.meta
         self.Tag = self._data.Tag
+        self.lang = self.meta.get('language_code')
 
         # extra attributes
         self.path = path
-        self.ee = self.words.compile_replaces({'е': 'ё'})
 
     def build_tag_info(self, para_id, idx):
         """
@@ -71,7 +70,7 @@ class Dictionary(object):
         Build a normal form.
         """
 
-        if idx == 0: # a shortcut: normal form is a word itself
+        if idx == 0:  # a shortcut: normal form is a word itself
             return fixed_word
 
         paradigm = self.paradigms[para_id]
@@ -104,11 +103,14 @@ class Dictionary(object):
         else:
             return fixed_word[len(prefix):]
 
-    def word_is_known(self, word, strict_ee=False):
+    def word_is_known(self, word, substitutes_compiled=None):
         """
         Check if a ``word`` is in the dictionary.
-        Pass ``strict_ee=True`` if ``word`` is guaranteed to
-        have correct е/ё letters.
+
+        To allow some fuzzyness pass ``substitutes_compiled`` argument;
+        it should be a result of :meth:`DAWG.compile_replaces()`.
+        This way you can e.g. handle ё letters replaced with е in the
+        input words.
 
         .. note::
 
@@ -118,10 +120,10 @@ class Dictionary(object):
             method should be used with extra care.
 
         """
-        if strict_ee:
-            return word in self.words
+        if substitutes_compiled:
+            return bool(self.words.similar_keys(word, substitutes_compiled))
         else:
-            return bool(self.words.similar_keys(word, self.ee))
+            return word in self.words
 
     def iter_known_words(self, prefix=""):
         """
@@ -135,7 +137,6 @@ class Dictionary(object):
             normal_form = self.build_normal_form(para_id, idx, word)
             yield word, tag, normal_form, para_id, idx
 
-
     def __repr__(self):
         return str("<%s>") % self.__class__.__name__
 
